local ARP = LibStub("AceAddon-3.0"):NewAddon("ARP", "AceTimer-3.0", "AceComm-3.0")
local libS = LibStub("AceSerializer-3.0")
local libC = LibStub("LibCompress")
local libCE = libC:GetAddonEncodeTable()

-- TODO: restructure the way assigning loot works:
--  Maintain master list of loot, with who it's assigned to, and quantity remaining
--   Adjust loot assignment window to accomodate the assignee info (in parentheses right of item name)
--  As each bid finishes, remove quantity from the main list, remove entirely when last one is bid out
--  New Unassign button
--  Move Reload to console command
--  If someone with loot assigned to them becomes ready (ie /arp reset, etc), do an Unassign
-- TODO: implement all new com msgs (LOOT, BIDLIST, BIDDING, BIDDING_ACK, BID, BID_ACK, REQUEST_STATUS)
--  Include 1-second throttling for all LOOT/BIDLIST msgs

-- Get release version and SVN revision into variables
local ARPVersion = 'r120'
local ARPRevision = '120'

-- Keybind labels (for keybinds defined in bindings.xml)
BINDING_HEADER_ARP		= "Angry Raid Points"
BINDING_NAME_ARP_TOGGLE = "Toggle Window"

-- Configurable variables
local rankMember = 4          -- lowest guild rank of Members
local rankMemberException = 2 -- exception to the above (in case you have an "officer alt" rank, if none, just set to -1)
local rankTrial = 5           -- guild rank of Trials
local minBidMMS = 10          -- minimum bid for members, mainspec
local minBidMOS = 1           -- minimum bid for members, offspec
local minBidTMS = 10          -- minimum bid for trials, mainspec
local minBidTOS = 1           -- minimum bid for trials, offspec
local maxLootLines = 150      -- maximum number of lines retained in the loot results log
-- Any items with these ilvls will have no suffix in the loot log:
local normalLoot = { 522, }
-- Any items with these ilvls will be suffixed with (H) in the loot log:
local heroicLoot = { 535, }	
-- Any other items will be suffixed with (ilvl) in the loot log.

-- Variables tied to loot assignment (assigning items to bid handlers)
local bCorePanelExtended = false  -- Is the loot assignment portion of the window shown?
local lastLootSeen = {}           -- Used to detect whether we've opened a new corpse (list of item links)
local lootList = {}               -- { {itemLink = ..., count = ...}, ... }
local assignList = {}             -- { {name = ...}, ... }
local lootBtnList = {}            -- { {btn = ...}, ... }
local assignBtnList = {}          -- { {btn = ...}, ... }
local selectedItem = nil          -- Which item is currently selected in the Loot Assignment UI
local selectedAssign = nil        -- Which bid handler is currently selected in the Loot Assignment UI

-- Variables tied to bid handling (handling the bidding of a single item)
local bBidding = false            -- Is bidding in progress yet?
local item = nil                  -- Item link of the item I'm handling
local bMult = false               -- More than one of this item?
local itemCount = 1               --  If so, this is the total number
local curNum = 1                  --  And this is the number of the one I'm currently bidding out
local memberMain = {}             -- These are the four bidding lists, format: {name = ..., points = ..., randnum = ...}
local memberOff = {}	
local trialMain = {}	
local trialOff = {}
local randnums = {}               -- Random numbers (to break bid list positioning ties) generated for players
local topBidder = nil             -- Name of top bidder
local currentBidder = nil         -- Name of current bidder
local bidderIndex = nil           -- Numeric index (in current tier's list) of current bidder
local bidCategory                 -- Current tier that's bidding (mm, mo, tm, to)
local minBid                      -- Minimum amount that must be bid by current bidder
local channel = "RAID"            -- Which channel is bid being conducted on? Configurable by user

-- Other misc variables
local bWarnedOOD = false          -- Has the user been warned of an out of date version yet?
local lastReq = 0                 -- Controls the sending of regular QUERY's, for a loot assigner
local versionList = {}            -- Collect the version numbers everyone in the guild is running

-- GLOBALS: LibStub
-- GLOBALS: BINDING_HEADER_ARP
-- GLOBALS: BINDING_NAME_ARP_TOGGLE
-- GLOBALS: ARP_Scale
-- GLOBALS: ARP_LootLog
-- GLOBALS: ARP_Toggle

-- Format for our addon communication (prefix is always "ARP")...
--
-- { "LOOT",
--   {
--     { [Item Link], [Quantity], [Handler or nil] },
--     ...
--   }
-- }
-- 
-- { "BIDLIST",
--   [Item Link],
--   { // Member MS tier
--     { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--     ...
--   },
--   { // Member OS tier
--     { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--     ...
--   },
--   { // Trial MS tier
--     { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--     ...
--   },
--   { // Trial OS tier
--     { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--     ...
--   },
--   [State, 0=bidding not started, 1=last call announced, 2=Member MS tier bidding, 3=Member OS tier bidding, 4=Trial MS tier bidding, 5=Trial OS tier bidding, 6=finished],
--   [HighBidder or nil],
--   [HighBidAmount or 0],
--   [CurrentBidder (numeric index) or nil if not started],
--   [Total Quantity],
--   [Number of item currently being bid, always 1 if Quantity=1 for this item, or if bidding hasn't started yet]
-- }
-- 
-- { "BIDDING", [Item Link], [Current Points, or nil to remove yourself], [MS? true or false] }
-- 
-- { "BIDDING_ACK", [Item Link], [Error text if there is an error, or nil] }
-- 
-- { "BID", [Item Link], [Bid Amount, or 0 to pass] }
-- 
-- { "BID_ACK", [Item Link], [Error text if there is an error, or nil] }
-- 
-- { "WINNER", [Item Link], [Number of the one won, 1 if there's only 1], [Player, or nil if DE'ed], [Winning Bid Amount, or nil if DE'ed], [Loot Log Text] }
-- 
-- { "REQUEST_STATUS" }
-- 
-- { "VER_QUERY" }
-- { "VERSION", [Version], [SVN Revision #] }
-- { "ASSIGN", [Item Link], [Quantity] }
-- { "QUERY" }
-- { "UPDATE" }
-- { "READY" }
-- { "NOT_READY" }

-- Constants for dealing with our addon communication
local COMMAND = 1

local LOOT_Items = 2
local LOOT_Items_Item_Link = 1
local LOOT_Items_Quantity = 2
local LOOT_Items_Handler = 3

local BIDLIST_Item_Link = 2
local BIDLIST_Bidders = 3
local BIDLIST_Bidders_MS = 1
local BIDLIST_Bidders_OS = 2
local BIDLIST_Bidders_MSTrial = 3
local BIDLIST_Bidders_OSTrial = 4
local BIDLIST_Bidders_Player = 1
local BIDLIST_Bidders_Points = 2
local BIDLIST_Bidders_Bid = 3
local BIDLIST_State = 4
local BIDLIST_HighBidder = 5
local BIDLIST_HighBidAmount = 6
local BIDLIST_CurrentBidder = 7
local BIDLIST_Quantity = 8
local BIDLIST_Number = 9

local BIDDING_Item_Link = 2
local BIDDING_Points = 3
local BIDDING_MS = 4

local BIDDING_ACK_Item_Link = 2
local BIDDING_ACK_Error = 3

local BID_Item_Link = 2
local BID_Bid = 3

local BID_ACK_Item_Link = 2
local BID_ACK_Error = 3

local WINNER_Item_Link = 2
local WINNER_Number = 3
local WINNER_Player = 4
local WINNER_Bid = 5
local WINNER_Loot_Log_Text = 6

local VERSION_Version = 2
local VERSION_SVN_Revision = 3

local ASSIGN_Item_Link = 2
local ASSIGN_Quantity = 3

-- Begin functions

function ARP:OnInitialize()
	-- Init Saved Variables (first time addon loads)
	if (ARP_Scale == nil) then ARP_Scale = 1 end
	if (ARP_LootLog == nil) then ARP_LootLog = {} end
	
	-- Slash Commands
	local options = {
		name = "ARP",
		handler = ARP,
		type = "group",
		args = {
			toggle = {
				type = "execute",
				name = "Toggle Window",
				desc = "Shows/hides the main window (also available in game keybindings)",
				func = function() ARP_Toggle() end
			},
			reset = {
				type = "execute",
				name = "Value Reset",
				desc = "Resets, cancelling any bid you're currently handling",
				func = function() ARP:Reset() end
			},
			lastcall = {
				type = "execute",
				name = "Last Call",
				desc = "Provides a last call alert for your assigned item",
				func = function() 
					if item then ARP:LastCall() 
					else print("You have not been assigned an item!") end
				end
			},
			bid = {
				type = "execute",
				name = "Start Bidding",
				desc = "Starts bidding for your assigned item",
				func = function() 
					if item then ARP:StartBid_UI()
					else print("You have not been assigned an item!") end 
				end
			},
			channel = {
				type = "select",
				name = "Channel Select",
				desc = function() return "Set channel in which to display bid information (current:|cFFFF0000 "..channel.."|r)" end,
				values = {raid="Raid", guild="Guild"},
				set = function(info, val) channel = strupper(val) print("Output channel changed to "..channel) end,
				get = function(info) return channel end,
				style = "dropdown"
			},	
			test = {
				type = "execute",
				name = "Test Loot",
				desc = "Loads test loot into the Loot Assignment window",
				func = function() 
					-- These two items are pre-cached elsewhere
					table.insert(lootList, {itemLink = select(2, GetItemInfo(19019)), count = 2})
					table.insert(lootList, {itemLink = select(2, GetItemInfo(87166)), count = 1})
					if (bCorePanelExtended) then ARP:MLPanel_Update() end
					print("Test loot loaded") 
				end
			},		
			scale = {
				type = "range",
				name = "Scale",
				desc = function() 
					return "Sets the scale of the ARP windows (current:|cFFFF0000 "..ARP_Scale.."|r)" 
				end,
				min = 0.3,
				max = 3,
				get = function(info) return ARP_Scale end,
				set = function(info, val) ARP_Scale = val end
			},
			version = {
				type = "execute",
				name = "Get Versions",
				desc = "Displays a list of all users (in the guild) running the addon and the version they're running",
				func = function() 
					versionList = {} -- start with a fresh version list, when displaying it
					ARP:SendMessage({ "VER_QUERY" }, "GUILD") 
					self:ScheduleTimer("VersionCheckOutput", 2)
					print("|cFF33FF99Angry Raid Points version check running...|r ")
				end
			},
			reminder = {
				type = "execute",
				name = "Remind Bidder",
				desc = "Resends tell to the current bidder",
				func = function() 
					if (bBidding) then SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
					else print("You are not currently bidding an item!") end
				end
			},
			remove = {
				type = "input",
				name = "Remove Bidder",
				desc = "Removes a bidder from the bidding list",
				set = function(info, val) 
					if (not item) then
						print ("Cannot remove bidders without an item to bid!")
					elseif (bBidding) then
						print ("Cannot remove bidders after bidding has started!")
					else
						ARP:RemoveBidder(val) 
					end
				end,
				get = function(info) return end
			}
		},
	}

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Angry Raid Points", options, {"arp"})
end

function ARP:OnEnable()
	local arpFrame = CreateFrame("Frame")
	arpFrame:RegisterEvent("CHAT_MSG_WHISPER")
	arpFrame:RegisterEvent("CHAT_MSG_BN_WHISPER")
	arpFrame:RegisterEvent("LOOT_OPENED")
	arpFrame:RegisterEvent("PARTY_LOOT_METHOD_CHANGED")
	arpFrame:RegisterEvent("GROUP_ROSTER_UPDATE")
	arpFrame:RegisterEvent("PLAYER_FLAGS_CHANGED")
	ARP:RegisterComm("ARP", "ReceiveMessage")
	
	arpFrame:SetScript("OnEvent", ARP.OnEvent)
	ARP:CreateCoreFrame()
	
	-- Cache these two items for possible /arp test usage...
	local _ = GetItemInfo(19019)
	_ = GetItemInfo(87166)
	
	self:ScheduleTimer("AfterEnable", 5)
end

function ARP:ReceiveMessage(prefix, data, channel, sender)
	if prefix ~= "ARP" then return end

	local one = libCE:Decode(data) -- Decode the compressed data
	local two, message = libC:Decompress(one) -- Decompress the decoded data
	if not two then error("Error decompressing: " .. message); return end
	local success, final = libS:Deserialize(two) -- Deserialize the decompressed data
	if not success then error("Error deserializing " .. final); return end

	ARP:ProcessMessage(sender, final)
end

function ARP:VersionCheckOutput()
	local versionliststr = ""
	for i,v in pairs(versionList) do
		versionliststr = versionliststr..v["name"].."-|cFFFF0000"..v["version"].."|r "
	end
	print(versionliststr) 
end

function ARP:AfterEnable()
	ARP:SendMessage({ "VER_QUERY" }, "GUILD")
	if (IsInRaid(LE_PARTY_CATEGORY_HOME)) then ARP:SendMessage({ "READY" }) end
end

function ARP:CopyLootLog()
	local lootLog = ARP.corePanel.lootLog.text
	local copyFrame = ARP.corePanel.lootLog.copyFrame.text
	local copyScrollFrame = ARP.corePanel.lootLog.copyFrame.scrollFrame
	local firstline = true
	
	copyFrame:SetText("")
	for i=1, lootLog:GetNumMessages(), 1 do
		local msg = lootLog:GetMessageInfo(i)
		if (strmatch(msg, "= DE$") == nil) then
			if firstline then 
				firstline = false
			else
				copyFrame:Insert("\n")
			end
			copyFrame:Insert( msg )
		end
	end

	copyScrollFrame:GetParent():Show()
	copyFrame:SetFocus()
	copyFrame:HighlightText(0)
end

function ARP:Set_LootLog_Scroll_Limits()
	local curLines = ARP.corePanel.lootLog.text:GetNumMessages()
	local maxScroll = maxLootLines - 1
	local minScroll = maxLootLines - curLines + 12
	if (minScroll > maxScroll) then minScroll = maxScroll end
	ARP.corePanel.lootLog.slider:SetMinMaxValues(minScroll, maxScroll)
	if (ARP.corePanel.lootLog.slider:GetValue() > maxScroll) then
		ARP.corePanel.lootLog.slider:SetValue(maxScroll)
	elseif (ARP.corePanel.lootLog.slider:GetValue() < minScroll) then
		ARP.corePanel.lootLog.slider:SetValue(minScroll)
	end
end

function ARP:CreateCoreFrame()
	local frame
	local button
	local title
	local text
	local dropdown
	local label	-- Predeclare variables to make future editing easier

	-- Core Panel
	frame = CreateFrame("Frame", "ARPCorePanel", UIParent)
	frame:SetSize(475, 350)
	frame:SetScale(ARP_Scale)
	frame:SetPoint("CENTER")
	frame:SetFrameStrata("HIGH")
	frame:SetFrameLevel(1)
	frame:SetBackdrop( {	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
							edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
							tile = true,
							tileSize = 32,
							edgeSize = 24,
							insets = { left = 5, right = 6, top = 6, bottom = 5 } } )
	frame:EnableMouse(true)
	title = frame:CreateTitleRegion()
	title:SetAllPoints()
	frame:SetScript("OnShow", ARP.CorePanel_OnShow)
	frame:SetScript("OnUpdate", ARP.CorePanel_OnUpdate)
	tinsert(UISpecialFrames, "ARPCorePanel")
	frame:Hide()
	ARP.corePanel = frame
	
	-- Core Panel Title Texture
	local texture = ARP.corePanel:CreateTexture()
	texture:SetTexture("Interface\\Addons\\AngryRaidPoints\\Angry_Logo.tga")
	texture:SetPoint("TOP", ARP.corePanel, "TOPLEFT", 220, 0)
	
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY")
	label:SetPoint("LEFT", texture, "RIGHT", 0, 5)
	label:SetJustifyH("LEFT")
	label:SetFont("Fonts\\FRIZQT__.TTF", 18, "OUTLINE")
	label:SetTextColor(1, 0.82, 0, 1)
	label:SetText("RP")
	
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	label:SetPoint("TOPLEFT", texture, "BOTTOMRIGHT", -15, 20)
	label:SetJustifyH("LEFT")
	local versiontext = ARPVersion
	if (versiontext:sub(1,1) == "@") then versiontext = "dev" end
	label:SetText(versiontext)
	
	-- Assigned Item Label
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	label:SetPoint("LEFT", ARP.corePanel, "LEFT", 10, 100)
	label:SetJustifyH("LEFT")
	label:SetText("ITEM")
	ARP.corePanel.itemLabel = label
	
	-- Assigned Item Text
	text = ARP.corePanel:CreateFontString(nil, "OVERLAY")
	text:SetPoint("LEFT", ARP.corePanel.itemLabel, "LEFT", 5, -20)
	text:SetJustifyH("LEFT")
	text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	ARP.corePanel.itemText = text
	
	-- Channel Output Label
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	label:SetPoint("LEFT", ARP.corePanel, "LEFT", 10, 50)
	label:SetJustifyH("LEFT")
	label:SetText("CHANNEL OUTPUT")
	ARP.corePanel.channelLabel = label
	
	-- Channel Output Dropdown
	dropdown = CreateFrame("Frame", "ChannelOutputDropdown", ARP.corePanel, "UIDropDownMenuTemplate")
	dropdown:SetPoint("LEFT", ARP.corePanel.channelLabel, "LEFT", -10, -25)
	ARP.corePanel.channelDropdown = dropdown
	
	-- Current Bidder Label
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	label:SetPoint("LEFT", ARP.corePanel, "LEFT", 10, 0)
	label:SetJustifyH("LEFT")
	label:SetText("CURRENT BIDDER")
	ARP.corePanel.curBidderLabel = label
	
	-- Current Bidder Text
	text = ARP.corePanel:CreateFontString(nil, "OVERLAY")
	text:SetPoint("LEFT", ARP.corePanel.curBidderLabel, "LEFT", 5, -20)
	text:SetJustifyH("LEFT")
	text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	ARP.corePanel.curBidderText = text
	
	-- Top Bidder Label
	label = ARP.corePanel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	label:SetPoint("LEFT", ARP.corePanel, "LEFT", 10, -50)
	label:SetJustifyH("LEFT")
	label:SetText("TOP BIDDER")
	ARP.corePanel.topBidderLabel = label
	
	-- Top Bidder Text
	text = ARP.corePanel:CreateFontString(nil, "OVERLAY")
	text:SetPoint("LEFT", ARP.corePanel.topBidderLabel, "LEFT", 5, -20)
	text:SetJustifyH("LEFT")
	text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	ARP.corePanel.topBidderText = text
	
	-- Core Frame Close Button
	button = CreateFrame("Button", "CoreCloseBtn", ARP.corePanel, "UIPanelCloseButton")
	button:SetPoint("TOPRIGHT", ARP.corePanel, "TOPRIGHT")
	ARP.corePanel.closeBtn = button
	
	-- Core Frame Start Bid Button
	button = CreateFrame("Button", "CoreBidBtn", ARP.corePanel, "UIPanelButtonTemplate")
	button:SetSize(125, 30)
	button:SetPoint("BOTTOM", ARP.corePanel, "BOTTOMLEFT", 235, 10)
	button:SetText("Start Bid")
	button:Disable()
	button:SetScript("OnClick", function()
		ARP:StartBid_UI()
	end)
	ARP.corePanel.startBidBtn = button
	
	-- Core Frame Loot Assignment Button
	button = CreateFrame("Button", "CoreMLPanelShowBtn", ARP.corePanel, "UIPanelButtonTemplate")
	button:SetSize(125, 30)
	button:SetPoint("BOTTOMRIGHT", ARP.corePanel, "BOTTOMLEFT", 460, 10)
	button:SetText("Loot Assignment")
	button:Disable()
	button:SetScript("OnClick", function()
		if (bCorePanelExtended) then 
			bCorePanelExtended = false
		else 
			bCorePanelExtended = true 
			ARP:MLPanel_Update()
		end
	end)
	ARP.corePanel.lootAssignBtn = button
	
	-- Core Frame Last Call Button
	button = CreateFrame("Button", "CoreBidListBtn", ARP.corePanel, "UIPanelButtonTemplate")
	button:SetSize(125, 30)
	button:SetPoint("BOTTOMLEFT", ARP.corePanel, "BOTTOMLEFT", 10, 10)
	button:SetText("Last Call")
	button:Disable()
	button:SetScript("OnClick", function()
		ARP:LastCall()
	end)
	ARP.corePanel.lastCallBtn = button
	
	-- Core Frame Reminder Button
	button = CreateFrame("Button", "CoreReminderBtn", ARP.corePanel, "UIPanelButtonTemplate")
	button:SetSize(125, 30)
	button:SetPoint("BOTTOM", ARP.corePanel.lastCallBtn, "TOP", 0, 0)
	button:SetText("Reminder")
	button:Disable()
	button:SetScript("OnClick", function()
		SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)			
	end)
	ARP.corePanel.reminderBtn = button
	
	-- Loot Log Container
	frame = CreateFrame("Frame", "CoreLootLogContainer", ARP.corePanel)
	frame:SetSize(280, 150)
	frame:SetPoint("RIGHT", ARP.corePanel, "LEFT", 440, -20)
	frame:SetBackdrop( {	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
							edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
							tile = true,
							tileSize = 32,
							edgeSize = 24,
							insets = { left = 5, right = 6, top = 6, bottom = 5 } } )
	ARP.corePanel.lootLog = frame
	
	-- Loot Log Label Text
	label = ARP.corePanel.lootLog:CreateFontString("LootLogLabel", "OVERLAY", "GameFontNormal")
	label:SetPoint("BOTTOM", ARP.corePanel.lootLog, "TOP")
	label:SetText("LOOT LOG")
	ARP.corePanel.lootLog.label = label
	
	frame = CreateFrame("ScrollingMessageFrame", "LootLogText", ARP.corePanel.lootLog)
	frame:SetSize(264, 134)
	frame:SetPoint("TOPLEFT", ARP.corePanel.lootLog, "TOPLEFT", 8, -8)
	frame:SetFadeDuration(0)
	frame:SetFading(false)
	frame:SetJustifyH("LEFT")
	frame:SetInsertMode("BOTTOM")
	frame:SetFontObject("GameFontNormalSmall")
	frame:SetMaxLines(maxLootLines) 
	ARP.corePanel.lootLog.text = frame
	
	frame = CreateFrame("Slider", "LootLogSlider", ARP.corePanel.lootLog.text, "UIPanelScrollBarTemplate")
	frame:SetBackdrop({
		bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
		edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
		tile = true, tileSize = 8, edgeSize = 8,
		insets = { left = 2, right = 2, top = 3, bottom = 3 }})
	frame:SetPoint("TOPLEFT", ARP.corePanel.lootLog, "TOPRIGHT", -2, -24)
	frame:SetPoint("BOTTOMLEFT", ARP.corePanel.lootLog, "BOTTOMRIGHT", -2, 24)
	frame:SetMinMaxValues(0, maxLootLines-1)
	frame:SetValueStep(1)
	frame:SetWidth(16)
	frame:SetScript("OnValueChanged", function(self)
		local newval = self:GetValue()
		local min
		local max
		min, max = self:GetMinMaxValues()
		if ( newval <= min ) then
					_G["LootLogSliderScrollUpButton"]:Disable()
		else
					_G["LootLogSliderScrollUpButton"]:Enable()
		end
		if ( newval >= max ) then
					_G["LootLogSliderScrollDownButton"]:Disable()
		else
					_G["LootLogSliderScrollDownButton"]:Enable()
		end
		ARP.corePanel.lootLog.text:SetScrollOffset(maxLootLines - 1 - newval)
	end)
	frame:SetScript("OnMinMaxChanged", function(self)
		local newval = self:GetValue()
		local min
		local max
		min, max = self:GetMinMaxValues()
		if ( newval <= min ) then
					_G["LootLogSliderScrollUpButton"]:Disable()
		else
					_G["LootLogSliderScrollUpButton"]:Enable()
		end
		if ( newval >= max ) then
					_G["LootLogSliderScrollDownButton"]:Disable()
		else
					_G["LootLogSliderScrollDownButton"]:Enable()
		end
	end)
	frame:SetValue(maxLootLines - 1)
	ARP.corePanel.lootLog.slider = frame
	
	ARP.corePanel.lootLog.text:EnableMouseWheel(true)
	ARP.corePanel.lootLog.text:SetScript("OnMouseWheel", function(self, delta)
	local current = ARP.corePanel.lootLog.slider:GetValue()
	if (delta < 0) then
			current = current + 3
			if (current > maxLootLines - 1) then
				current = maxLootLines - 1
			end
			ARP.corePanel.lootLog.slider:SetValue(current) 
	elseif (delta > 0) then
			current = current - 3
			if (current < 0) then
				current = 0
			end
			ARP.corePanel.lootLog.slider:SetValue(current)
	end
	end)
	
	if (#ARP_LootLog) then -- load in saved loot log
		while (#ARP_LootLog > maxLootLines) do  -- if we decrease the amount of loot to store later on, handle it properly
			tremove(ARP_LootLog, 1)
		end
		for i=1,#ARP_LootLog,1 do
			ARP.corePanel.lootLog.text:AddMessage(ARP_LootLog[i])
		end
	end
	
	ARP:Set_LootLog_Scroll_Limits()

	-- Loot Log Copy Container
	frame = CreateFrame("Frame", "LootLogCopyFrame", ARP.corePanel.lootLog)
	frame:SetSize(280, 150)
	frame:SetPoint("CENTER", ARP.corePanel, "TOP", 0, 0)
	frame:SetFrameLevel(ARP.corePanel.lootLog:GetFrameLevel()+2)
	frame:SetBackdrop( {	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
							edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
							tile = true,
							tileSize = 32,
							edgeSize = 24,
							insets = { left = 5, right = 6, top = 6, bottom = 5 } } )
	title = frame:CreateTitleRegion()
	title:SetAllPoints()
	frame:Hide()
	ARP.corePanel.lootLog.copyFrame = frame
	
	-- Loot Log Copy Container Scroll Frame
	frame = CreateFrame("ScrollFrame", "LootLogCopyScrollFrame", ARP.corePanel.lootLog.copyFrame, "UIPanelScrollFrameTemplate")
	frame:SetPoint("TOPLEFT", ARP.corePanel.lootLog.copyFrame, "TOPLEFT", 8, -8)
	frame:SetPoint("BOTTOMRIGHT", ARP.corePanel.lootLog.copyFrame, "BOTTOMRIGHT", -8, 8)
	ARP.corePanel.lootLog.copyFrame.scrollFrame = frame
	
	ARP.corePanel.lootLog.copyFrame.scrollFrame.scrollBar = _G["LootLogCopyScrollFrameScrollBar"]
	ARP.corePanel.lootLog.copyFrame.scrollFrame.scrollBar:SetBackdrop({
		bgFile = "Interface\\Buttons\\UI-SliderBar-Background",
		edgeFile = "Interface\\Buttons\\UI-SliderBar-Border",
		tile = true, tileSize = 8, edgeSize = 8,
		insets = { left = 2, right = 2, top = 3, bottom = 3 }})
	
	-- Loot Log Copy Container Text
	frame = CreateFrame("EditBox", "LootLogCopyFrameText", ARP.corePanel.lootLog.copyFrame)
	ARP.corePanel.lootLog.copyFrame.scrollFrame:SetScrollChild(frame)
	frame:SetAutoFocus(false)
	frame:SetMultiLine(true)
	frame:SetMaxLetters(0)
	frame:EnableMouse(true) -- moved this from main frame to editbox
	frame:SetFontObject("GameFontNormalSmall")	
	frame:SetWidth(264)
	frame:SetScript("OnUpdate", function(self)
		ScrollingEdit_OnUpdate(self)
	end)
	frame:SetScript("OnTextChanged", function(self)
		ScrollingEdit_OnTextChanged(self, self:GetParent())
	end)
	frame:SetScript("OnEscapePressed", function(self)
		self:ClearFocus()
	end)
	frame:SetScript("OnCursorChanged", function(self, x, y, width, height)
		ScrollingEdit_OnCursorChanged(self, x, y, width, height)
	end)
	ARP.corePanel.lootLog.copyFrame.text = frame

	ARP.corePanel.lootLog.copyFrame.scrollFrame:UpdateScrollChildRect()
		
	-- Loot Log Copy Container Close Button
	button = CreateFrame("Button", "CopyContainerCloseBtn", ARP.corePanel.lootLog.copyFrame, "UIPanelButtonTemplate")
	button:SetSize(150, 30)
	button:SetPoint("CENTER", ARP.corePanel.lootLog.copyFrame, "BOTTOM", 0, -15)
	button:SetText("Close")
	button:SetScript("OnClick", function()
		ARP.corePanel.lootLog.copyFrame:Hide()
	end)
	
	-- Copy Loot Log Button
	button = CreateFrame("Button", "CopyLootLogBtn", ARP.corePanel.lootLog, "UIPanelButtonTemplate")
	button:SetSize(75, 30)
	button:SetPoint("CENTER", ARP.corePanel.lootLog, "BOTTOM", -75, -15)
	button:SetText("Copy")
	button:SetScript("OnClick", function()
		if (not ARP.corePanel.lootLog.copyFrame:IsShown()) then
			ARP:CopyLootLog()
		end
	end)
	ARP.corePanel.lootLog.copyBtn = button
	
	-- Clear Loot Log Button
	button = CreateFrame("Button", "ClearLootLogBtn", ARP.corePanel.lootLog, "UIPanelButtonTemplate")
	button:SetSize(75, 30)
	button:SetPoint("CENTER", ARP.corePanel.lootLog, "BOTTOM", 75, -15)
	button:SetText("Clear")
	button:SetScript("OnClick", function()
		ARP.corePanel.lootLog.text:Clear()
		ARP_LootLog = {}
		ARP:Set_LootLog_Scroll_Limits()
	end)
	ARP.corePanel.lootLog.clearBtn = button
	
	-----------------------------)
	---Loot Assignment Portion---)
	-----------------------------)
	
	-- Loot Assignment Item List Container
	frame = CreateFrame("Frame", "MLItemList", ARP.corePanel)
	frame:SetSize(300, 120)
	frame:SetPoint("TOP", ARP.corePanel, "TOPRIGHT", -187.5, -20)
	frame:SetBackdrop( {	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
							edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
							tile = true,
							tileSize = 32,
							edgeSize = 24,
							insets = { left = 5, right = 6, top = 6, bottom = 5 } } )
	frame:Hide()
	ARP.corePanel.itemList = frame
	
	-- Loot Assignment Item List Label
	label = ARP.corePanel.itemList:CreateFontString("ItemListLabel", "OVERLAY", "GameFontNormal")
	label:SetPoint("BOTTOM", ARP.corePanel.itemList, "TOP")
	label:SetJustifyH("CENTER")
	label:SetText("ITEM LIST")
	ARP.corePanel.itemList.label = label
	
	-- Loot Assignment Item List Reload Button
	button = CreateFrame("Button", "MLLootClearBtn", ARP.corePanel.itemList, "UIPanelButtonTemplate")
	button:SetSize(75, 30)
	button:SetPoint("TOP", ARP.corePanel.itemList, "BOTTOM", 0, 0)
	button:SetText("Reload")
	button:SetScript("OnClick", function()
		ARP:MLPanel_ClearLoot()
	end)
	ARP.corePanel.itemList.clearBtn = button
	
	-- Loot Assignment Assignment List
	frame = CreateFrame("Frame", "MLAssignList", ARP.corePanel)
	frame:SetSize(150, 120)
	frame:SetPoint("BOTTOM", ARP.corePanel, "BOTTOMRIGHT", -187.5, 40)
	frame:SetBackdrop( {	bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",
							edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
							tile = true,
							tileSize = 32,
							edgeSize = 24,
							insets = { left = 5, right = 6, top = 6, bottom = 5 } } )
	frame:Hide()
	ARP.corePanel.assignList = frame
	
	-- Loot Assignment Assignment List Label
	label = ARP.corePanel.assignList:CreateFontString("AssignListLabel", "OVERLAY", "GameFontNormal")
	label:SetPoint("BOTTOM", ARP.corePanel.assignList, "TOP")
	label:SetJustifyH("CENTER")
	label:SetText("ASSIGN LIST")
	ARP.corePanel.assignList.label = label	
	
	-- Loot Assignment Assignment List Loot Assignment Button (Hahahaha)
	button = CreateFrame("Button", "MLAssignBtn", ARP.corePanel.assignList, "UIPanelButtonTemplate")
	button:SetSize(75, 30)
	button:SetPoint("TOP", ARP.corePanel.assignList, "BOTTOM", 0, 0)
	button:SetText("Assign")
	button:SetScript("OnClick", function()
		if (selectedItem and selectedAssign) then
			ARP:MLPanel_AssignLoot()
			ARP:MLPanel_Update()
		end
	end)
	ARP.corePanel.assignBtn = button
end

function ARP.CorePanel_OnShow(self)
	self:SetScale(ARP_Scale)
	if (itemCount > 1) then	ARP.corePanel.itemText:SetText(item.."x"..itemCount)
	else ARP.corePanel.itemText:SetText(item) end
	ARP.corePanel.curBidderText:SetText(currentBidder)
	if (minBid ~= nil and topBidder ~= nil) then
		ARP.corePanel.topBidderText:SetText(topBidder) 
	end
	
	local function itemOnClick(self)
		local selected = self.value
		channel = selected
		UIDropDownMenu_SetSelectedValue(ARP.corePanel.channelDropdown, selected)
	end
	UIDropDownMenu_Initialize(ARP.corePanel.channelDropdown, function()
		local info = UIDropDownMenu_CreateInfo()
		local channels = {"RAID", "GUILD"}
		for i, chnl in next, channels do
			info.text = channels[i]
			info.value = chnl
			info.func = itemOnClick
			info.checked = chnl == channel
			UIDropDownMenu_AddButton(info)
		end
	end)
	UIDropDownMenu_SetSelectedValue(ARP.corePanel.channelDropdown, channel)
	UIDropDownMenu_SetWidth(ARP.corePanel.channelDropdown, 80)
	UIDropDownMenu_JustifyText(ARP.corePanel.channelDropdown, "LEFT")
	
	if (item) then
		ARP.corePanel.startBidBtn:Enable()
		ARP.corePanel.lastCallBtn:Enable()
	else 
		ARP.corePanel.startBidBtn:Disable()
		ARP.corePanel.lastCallBtn:Disable()
	end
	
	ARP.corePanel:SetScript("OnShow", nil)	
end

function ARP.CorePanel_OnUpdate(self, elapsed)
	if (not bCorePanelExtended) then	
		ARP.corePanel:SetSize(470, 350)
		ARP.corePanel.itemList:Hide()
		ARP.corePanel.assignList:Hide()
	else
		ARP.corePanel:SetSize(810, 350)
		ARP.corePanel.itemList:Show()
		ARP.corePanel.assignList:Show()
	end
	
	self:SetScale(ARP_Scale)
	if (itemCount > 1) then ARP.corePanel.itemText:SetText(item.."x"..itemCount)
	else ARP.corePanel.itemText:SetText(item) end
	
	UIDropDownMenu_SetSelectedValue(ARP.corePanel.channelDropdown, channel)
	
	if (currentBidder ~= nil) then
		if (UnitName("player") == currentBidder or UnitInRaid(currentBidder)) then
			local color = RAID_CLASS_COLORS[strupper(UnitClass(currentBidder))]
			if color then
				ARP.corePanel.curBidderText:SetText("|c"..color.colorStr..currentBidder.."|r")
			else
				ARP.corePanel.curBidderText:SetText(currentBidder)
			end
		else 
			ARP.corePanel.curBidderText:SetText(currentBidder) 
		end
	else 
		ARP.corePanel.curBidderText:SetText(nil)
	end
	
	if (minBid ~= nil and topBidder ~= nil) then
		if (UnitName("player") == topBidder or UnitInRaid(topBidder)) then
			local color = RAID_CLASS_COLORS[strupper(UnitClass(topBidder))]
			if color then
				ARP.corePanel.topBidderText:SetText("|c"..color.colorStr..topBidder.."|r at "..(minBid-1).." points")
			else
				ARP.corePanel.topBidderText:SetText(topBidder.." at "..(minBid-1).." points")
			end
		else
			ARP.corePanel.topBidderText:SetText(topBidder.." at "..(minBid-1).." points") 
		end
	else 
		ARP.corePanel.topBidderText:SetText(nil)
	end
	
	-- Button enable/disable checks
	if (ARP.corePanel.lootLog.text:GetNumMessages() > 0) then
		ARP.corePanel.lootLog.copyBtn:Enable()
		ARP.corePanel.lootLog.clearBtn:Enable()
	else
		ARP.corePanel.lootLog.copyBtn:Disable()
		ARP.corePanel.lootLog.clearBtn:Disable()
	end
	
	if (item) then
		ARP.corePanel.startBidBtn:Enable()
		ARP.corePanel.lastCallBtn:Enable()
	else
		ARP.corePanel.startBidBtn:Disable()
		ARP.corePanel.lastCallBtn:Disable()
	end
	
	if (bBidding) then ARP.corePanel.reminderBtn:Enable()
	else ARP.corePanel.reminderBtn:Disable() end

	if (IsInRaid(LE_PARTY_CATEGORY_HOME)) then
		local mlName
		local method, _, mlID = GetLootMethod()
		if (method == "master") then mlName = GetRaidRosterInfo(mlID) end
		if ((method ~= "master") or (mlName == UnitName("player"))) then 
			ARP.corePanel.lootAssignBtn:Enable()
			-- Every 30 seconds when window is visible, re-check who's out there so we always have at least semi-updated info
			if (bCorePanelExtended and (lastReq < (GetTime() - 30))) then
				ARP:Cleanup_AssignList()
				ARP:SendMessage({ "QUERY" })
				lastReq = GetTime()
			end
		else 
			ARP.corePanel.lootAssignBtn:Disable() 
		end
	else 
		ARP.corePanel.lootAssignBtn:Disable()
	end
end

function ARP_Toggle()
	if ARP.corePanel:IsShown() then 
		ARP.corePanel:Hide() 
	else 
		ARP.corePanel:Show() 
	end
end

function ARP:MLPanel_Update()
	if (not bCorePanelExtended) then return end
	for i,v in pairs(lootBtnList) do v["btn"]:Hide() end
	lootBtnList = {}
	for i=1, #lootList do
		local button = CreateFrame("CheckButton", "LootButton"..i, ARP.corePanel.itemList, "UIRadioButtonTemplate")
		button.text = lootList[i]["itemLink"]
		button.count = lootList[i]["count"]
		local font = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		font:SetPoint("LEFT", button, "RIGHT", 5, 1)
		button:SetFontString(font)
		button:SetHitRectInsets(0, -225, 0, 0)
		if (selectedItem == i) then button:SetChecked(true) end
		if (lootList[i]["count"] > 1) then button:SetText(button.text.."x"..lootList[i]["count"])
		else button:SetText(button.text) end
		if (i > 1) then 
			button:SetPoint("TOPLEFT", ARP.corePanel.itemList.prevBtn, "BOTTOMLEFT", 0, 0)
		else button:SetPoint("TOPLEFT", ARP.corePanel.itemList, "TOPLEFT", 5, -5) end
		button:SetScript("OnEnter", function(...) 
				GameTooltip:SetOwner(ARP.corePanel.lastCallBtn, "ANCHOR_TOPRIGHT")
				GameTooltip:SetHyperlink(button.text)
				GameTooltip:Show()
				end)
		button:SetScript("OnLeave", function(...) GameTooltip:Hide() end)
		button:SetScript("OnClick", function(self, button, up)
			if IsModifiedClick() then
				-- Re-toggle the actual button then do the proper action
				local checked = self:GetChecked()
				self:SetChecked(not checked)
				HandleModifiedItemClick(self:GetText())
			end
			if (self:GetChecked()) then
				selectedItem = i
				ARP:MLPanel_SelectItem(i)
			else 
				selectedItem = nil
			end
		end)
		button:Show()
		ARP.corePanel.itemList.prevBtn = button
		table.insert(lootBtnList, {btn = button})
	end
	
	for i,v in pairs(assignBtnList) do v["btn"]:Hide() end
	assignBtnList = {}
	for i=1, #assignList do
		local button = CreateFrame("CheckButton", "AssignButton"..i, ARP.corePanel.assignList, "UIRadioButtonTemplate")
		button.text = assignList[i]["name"]
		if (UnitIsAFK(assignList[i]["name"])) then button.text = button.text.." (AFK)" end
		local font = button:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		font:SetPoint("LEFT", button, "RIGHT", 5, 1)
		button:SetFontString(font)
		button:SetHitRectInsets(0, -225, 0, 0)
		button:SetText(button.text)
		if (selectedAssign == i) then button:SetChecked(true) end
		if (i > 1) then	button:SetPoint("TOPLEFT", ARP.corePanel.assignList.prevBtn, "BOTTOMLEFT", 0, 0)
		else button:SetPoint("TOPLEFT", ARP.corePanel.assignList, "TOPLEFT", 5, -5) end
		button:SetScript("OnClick", function(self, button, up)
			if (self:GetChecked()) then
				selectedAssign = i
				ARP:MLPanel_SelectAssign(i)
			else 
				selectedAssign = nil 
			end
		end)
		button:Show()
		ARP.corePanel.assignList.prevBtn = button
		table.insert(assignBtnList, {btn = button})
	end

end

function ARP:MLPanel_AssignLoot()
	ARP:SendMessage({ "ASSIGN", lootBtnList[selectedItem]["btn"].text, lootBtnList[selectedItem]["btn"].count }, 
					"WHISPER", assignList[selectedAssign]["name"])
	table.remove(lootList, selectedItem)
	table.remove(assignList, selectedAssign)
	selectedItem = nil
	selectedAssign = nil
end

function ARP:MLPanel_ClearLoot()
	lootList = {}
	lastLootSeen = {}
	selectedItem = nil
	ARP:AttemptLootScan()
	ARP:MLPanel_Update()
end

function ARP:AttemptLootScan()
	local thisLootSeen = {}

	for i=1, GetNumLootItems() do
		if (GetLootSlotType(i) == 1) then
			local itemLink = GetLootSlotLink(i)
			local _, _, _, rarity = GetLootSlotInfo(i)
			local _ = GetItemInfo(itemLink) -- just to have cached for later
			if (rarity == 4) then table.insert(thisLootSeen, itemLink) end
		end
	end
	
	if (#thisLootSeen > 0) then
		local newloot = false
		local found = false
		
		-- If we find any loot here that wasn't in our last set, then we assume this is a new corpse/chest.
		for i=1, #thisLootSeen do
			found = false
			
			for _,v in pairs(lastLootSeen) do
				if (v == thisLootSeen[i]) then
					found = true
					break
				end
			end
		
			if found == false then
				newloot = true
				break
			end
		end
		
		if (newloot) then 
			ARP:Cleanup_AssignList()
			ARP:SendMessage({ "QUERY" })
			lastReq = GetTime()
			lastLootSeen = {}
			for i,v in pairs(thisLootSeen) do
				table.insert(lastLootSeen, v)
			end
			ARP:BuildItemList(thisLootSeen) 
		end
	end
end

function ARP:MLPanel_SelectItem(selected)
	for i,v in pairs(lootBtnList) do
		if (i ~= selected) then v["btn"]:SetChecked(false) end
	end
end

function ARP:MLPanel_SelectAssign(selected)
	for i,v in pairs(assignBtnList) do
		if (i ~= selected) then v["btn"]:SetChecked(false) end
	end
end

function ARP:SendMessage(data, channel, target)
	local one = libS:Serialize(data)
	local two = libC:CompressHuffman(one)
	local final = libCE:Encode(two)
	local destChannel = channel or "RAID"
	
	if destChannel ~= "RAID" or IsInRaid(LE_PARTY_CATEGORY_HOME) then -- don't try to send if not in a real raid ("You are not in a raid group" error)
		self:SendCommMessage("ARP", final, destChannel or "RAID", target, "NORMAL")
	end
end

function ARP:ProcessMessage(sender, data)
	local cmd = data[COMMAND]
	
	-- Requesting an update of everyone's master loot panel (to update AFK status or whatever)
	if (cmd == "UPDATE") then
		ARP:MLPanel_Update()
		
	-- Add a result to the loot log
	elseif (cmd == "WINNER") then
		ARP.corePanel.lootLog.text:AddMessage(data[WINNER_Loot_Log_Text])	
		tinsert(ARP_LootLog, data[WINNER_Loot_Log_Text])
		if (#ARP_LootLog > maxLootLines) then
			tremove(ARP_LootLog, 1)
		end
		ARP:Set_LootLog_Scroll_Limits()
			
	-- Assign us an item
	elseif (cmd == "ASSIGN") then
		local _ = GetItemInfo(data[ASSIGN_Item_Link]) -- just to have cached for later

		item = data[ASSIGN_Item_Link]
		itemCount = data[ASSIGN_Quantity]
		if (itemCount > 1) then bMult = true else bMult = false end

		ARP:SendMessage({ "NOT_READY" })

		if (itemCount > 1) then -- Let's always put the assignment messages on the raid channel?
			SendChatMessage(item.."x"..itemCount..": tells to "..UnitName("player"), "RAID")
		else 
			SendChatMessage(item..": tells to "..UnitName("player"), "RAID")
		end
			
	-- Someone's querying who's ready
	elseif (cmd == "QUERY") then
		if (not item) then ARP:SendMessage({ "READY" }) end
			
	-- Someone's broadcasting availability
	elseif (cmd == "READY") then
		local found = false
		local i = 1
		while i <= #assignList do
			if sender == assignList[i]["name"] then
				found = true
				if not UnitInRaid(sender) then
					if selectedAssign ~= nil and selectedAssign > i then
						selectedAssign = selectedAssign - 1
					elseif selectedAssign ~= nil and selectedAssign == i then
						selectedAssign = nil
					end
					table.remove(assignList, i)
				else
					i = i + 1
				end
			else
				i = i + 1
			end
		end
		if (not found) and UnitInRaid(sender) then 
			table.insert(assignList, {name = sender}) 
		end
		ARP:MLPanel_Update()
			
	-- Someone's broadcasting that they're no longer available
	elseif (cmd == "NOT_READY") then
		local found = false
		local i = 1
		while i <= #assignList do
			if sender == assignList[i]["name"] then
				found = true
				if selectedAssign ~= nil and selectedAssign > i then
					selectedAssign = selectedAssign - 1
				elseif selectedAssign ~= nil and selectedAssign == i then
					selectedAssign = nil
				end
				table.remove(assignList, i)
			else
				i = i + 1
			end
		end
		if found then ARP:MLPanel_Update() end
			
	-- Another user has requested this user's current version
	elseif (cmd == "VER_QUERY") then
		local revToSend
		local verToSend
		if ARPRevision:sub(1,1) == "@" then revToSend = "dev" else revToSend = tonumber(ARPRevision) end
		if ARPVersion:sub(1,1) == "@" then verToSend = "dev" else verToSend = ARPVersion end
		ARP:SendMessage({ "VERSION", verToSend, revToSend }, "GUILD")
			
	-- Received version info from other user
	elseif (cmd == "VERSION") then
		local localRev, rev, ver
		
		if ARPRevision:sub(1,1) == "@" then localRev = nil else localRev = tonumber(ARPRevision) end
		rev = data[VERSION_SVN_Revision]
		ver = data[VERSION_Version]
			
		if ((localRev ~= nil) and (rev > localRev) and (not bWarnedOOD)) then 
			print("|cFF33FF99Angry Raid Points|r: Your version of AngryRaidPoints is out of date! Download the latest version from www.wowace.com.")
			bWarnedOOD = true
		end

		local found = false
		for i,v in pairs(versionList) do
			if (v["name"] == sender) then
				v["version"] = ver
				found = true
			end
		end
		if (not found) then table.insert(versionList, {name = sender, version = ver}) end
	end
end

function ARP.OnEvent(self, event, ...)
	-- Whisper
	if (event == "CHAT_MSG_WHISPER") then
		local msg, sender = ...
		if (UnitInRaid(sender)) then ARP:HandleWhisper(msg, sender)
		else return end
		
	-- Battle.net Whisper
	elseif (event == "CHAT_MSG_BN_WHISPER") then
		local msg, sender, _, _, _, _, _, _, _, _, _, _, presenceID = ...
		local _, onlineBnetFriends = BNGetNumFriends()
		for i=1, onlineBnetFriends do
			local friendID, _, _, _, toonName, _, client = BNGetFriendInfo(i)
			if (client == BNET_CLIENT_WOW and presenceID == friendID) then
				if (UnitInRaid(toonName)) then ARP:HandleWhisper(msg, toonName) break
				else return end
			end
		end
		
	-- Player flags changed (AFK)
	elseif ( IsInRaid(LE_PARTY_CATEGORY_HOME) and (event == "PLAYER_FLAGS_CHANGED")) then
		local unitID = ...
		if (unitID == "player") then
			ARP:SendMessage({ "UPDATE" })
		end
		
	-- If player is the master looter (or if master loot isn't set), generate the loot into an assignable list
	elseif ( IsInRaid(LE_PARTY_CATEGORY_HOME) and (event == "LOOT_OPENED") ) then
		local mlName
		local method, _, mlID = GetLootMethod()
		if (method == "master") then mlName = GetRaidRosterInfo(mlID) end
		if ((method ~= "master") or (mlName == UnitName("player"))) then ARP:AttemptLootScan() end
	elseif ( IsInRaid(LE_PARTY_CATEGORY_HOME) and (event == "PARTY_LOOT_METHOD_CHANGED") ) then
		bCorePanelExtended = false
	elseif ( IsInRaid(LE_PARTY_CATEGORY_HOME) and (event == "GROUP_ROSTER_UPDATE") ) then
		ARP:Cleanup_AssignList()
	end
end

function ARP:Cleanup_AssignList() 
	local changes = false
	-- Clean up any outdated entries from assignList
	local al = 1
	while al <= #assignList do
		if (not UnitInRaid(assignList[al]["name"])) or (not UnitIsConnected(assignList[al]["name"])) then
			if selectedAssign ~= nil and selectedAssign > al then
				selectedAssign = selectedAssign - 1
			elseif selectedAssign ~= nil and selectedAssign == al then
				selectedAssign = nil
			end
			table.remove(assignList, al)
			changes = true
		else
			al = al + 1
		end
	end
	if changes then ARP:MLPanel_Update() end
end

function ARP:GetGuildRank(player)
	for i = 1, GetNumGuildMembers() do
		local name, _, rankIndex = GetGuildRosterInfo(i)
		if name and (name == player) then
			return rankIndex
		end
	end
	return -1
end

function ARP:HandleWhisper(msg, sender)
	msg = { strsplit(" ", strlower(msg)) }
	local points, spec = "", ""
	if (not bBidding) then 
		if (msg[1] == "remove") then ARP:RemoveBidder(sender) end
		if ( (strfind(msg[1], "%a+") and msg[2] == nil) or (strfind(msg[1], "%a+") and strfind(msg[2], "%a+")) ) then return end
		if ( strfind(msg[1], "^-*%d+$") and (msg[2] == "" or msg[2] == nil or strfind(msg[2], "^%a+$")) ) then
			points = msg[1]
			spec = msg[2]
		elseif ( strfind(msg[1], "^%a+$") and strfind(msg[2], "^-*%d+$") ) then
			points = msg[2]
			spec = msg[1]
		end
	else points = msg[1]
	end
	if (points == "nope" or points == "pass" or points == "yup" or strfind(points, "^-*%d+$") or 
		((points:sub(1,1) == "+") and (tonumber(points:sub(2)) > 0)))
		and item then
		if (bBidding) then
			if (sender == currentBidder) then
				local list = ARP:GetList(bidCategory)
				if (points ~= "nope" and points ~= "pass" and list[bidderIndex]["points"] < minBid) then
					ARP:UpdateBid(-1)
				else
					ARP:UpdateBid(points)
				end
			end
		else
			points = tonumber(points)
			local response
			
			-- Generalize spec
			if (spec == "ms" or spec == "" or spec == nil) then spec = "main"
			elseif (spec == "os") then spec = "off"
			end
			
			if (spec ~= "main" and spec ~= "off") then
				response = "Invalid spec! Please resend tell. Valid options are main/ms or off/os. You can also not mention a spec, to default to main spec."
			else
				local guildRankIndex = ARP:GetGuildRank(sender)
				if (guildRankIndex >= 0 and guildRankIndex <= rankMember and guildRankIndex ~= rankMemberException) then
					if (spec == "main") then ARP:UpdateList(memberMain, sender, points, memberOff)
					else ARP:UpdateList(memberOff, sender, points, memberMain) end
					response = "You're bidding on " .. item .. ": ["..points.."] points, ["..spec.."] spec."				
				elseif (guildRankIndex == rankTrial) then
					if (spec == "main") then ARP:UpdateList(trialMain, sender, points, trialOff)
					else ARP:UpdateList(trialOff, sender, points, trialMain) end
					response = "You're bidding on " .. item .. ": ["..points.."] points, ["..spec.."] spec."				
				elseif (guildRankIndex == -1) then
					response = "Unable to obtain your guild rank currently. Try resending your tell."
				else
					response = "Your guild rank is invalid. Tell ignored."
				end
			end
			if (sender ~= UnitName("player")) then SendChatMessage(response, "WHISPER", nil, sender)
			else print(response) end
		end
	end	
end

function ARP:BidItem(channel)
	local msg
	if (bMult) then msg = (item.." #"..curNum..": Starting bid! ")
	else msg = (item..": Starting bid! ") end
	local name, response
	
	memberMain = ARP:BidSort(memberMain)
	memberOff = ARP:BidSort(memberOff)
	
	if (memberMain[1]) then
		name = memberMain[1]["name"]
		minBid = minBidMMS
		bidCategory = "mm"
	elseif (memberOff[1]) then
		name = memberOff[1]["name"]
		minBid = minBidMOS
		bidCategory = "mo"
	elseif (trialMain[1]) then
		name = trialMain[1]["name"]
		minBid = minBidTMS
		bidCategory = "tm"
	elseif (trialOff[1]) then
		name = trialOff[1]["name"]
		minBid = minBidTOS
		bidCategory = "to"
	end
	
	SendChatMessage(msg..name.." min "..minBid, channel)
	SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, name)
	currentBidder = name
	bidderIndex = 1
	bBidding = true
end

function ARP:BuildBidString(itemLink)
	local response
	if (itemCount > 1) then response = (itemLink.."x"..itemCount..": ")
	else response = (itemLink..": ") end
		
	memberMain = ARP:BidSort(memberMain)
	for i,v in pairs(memberMain) do
		if (i == #memberMain) then response = (response..v["name"].." "..v["points"].." ")
		else response = (response..v["name"].." "..v["points"]..", ") end
	end
	memberOff = ARP:BidSort(memberOff)
	for i,v in pairs(memberOff) do
		if (i == 1) then response = (response.."(") end
		if (i == #memberOff) then response = (response..v["name"].." "..v["points"].." OS) ")
		else response = (response..v["name"].." "..v["points"]..", ") end
	end
	trialMain = ARP:BidSort(trialMain)
	for i,v in pairs(trialMain) do
		if (i == 1) then response = (response.."(") end
		if (i == #trialMain) then response = (response..v["name"].." "..v["points"].." Trial) ")
		else response = (response..v["name"].." "..v["points"]..", ") end
	end
	trialOff = ARP:BidSort(trialOff)
	for i,v in pairs(trialOff) do
		if (i == 1) then response = (response.."(") end
		if (i == #trialOff) then response = (response..v["name"].." "..v["points"].." Trial OS) ")
		else response = (response..v["name"].." "..v["points"]..", ") end
	end
	response = response.."LAST CALL"
	
	return response	
end

function ARP:BuildItemList(itemList)
	lootList = {}
	selectedItem = nil
	
	for i,v in pairs(itemList) do
		local found = false
		for x,y in pairs(lootList) do
			if (y["itemLink"] == v) then
				y["count"] = y["count"] + 1
				found = true
			end
		end
		if (not found) then table.insert(lootList, {itemLink = v, count = 1}) end
	end
	
	if (bCorePanelExtended) then ARP:MLPanel_Update() end
end

function ARP:LastCall()
	if (bBidding) then
		print("Your bid is already in progress!")
		return
	end
	if (item) then
		local response = ARP:BuildBidString(item)
		SendChatMessage(response, channel)
	end
end

function ARP:EndBid(winner, winAmt)
	if (winner == nil or winner == "None") then		-- Everyone passed
		if (bidCategory == "mm") then
			if (memberOff[1]) then
				bidCategory = "mo"
				bidderIndex = 1
				currentBidder = memberOff[1]["name"]
				minBid = minBidMOS
				SendChatMessage("Moving to member off specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			elseif (trialMain[1]) then
				bidCategory = "tm"
				bidderIndex = 1
				currentBidder = trialMain[1]["name"]
				minBid = minBidTMS
				SendChatMessage("Moving to trial main specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			elseif (trialOff[1]) then
				bidCategory = "to"
				bidderIndex = 1
				currentBidder = trialOff[1]["name"]
				minBid = minBidTOS
				SendChatMessage("Moving to trial off specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			else
				if (bMult) then SendChatMessage(item.." #"..curNum..": DE", channel)
				else SendChatMessage(item..": DE", channel) end
				ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
				if (itemCount > 1) then
					itemCount = itemCount - 1
					curNum = curNum + 1
					currentBidder = nil
					topBidder = nil
					ARP:StartBid()
				else
					ARP:Reset()
				end
			end
		elseif (bidCategory == "mo") then
			if (trialMain[1]) then
				bidCategory = "tm"
				bidderIndex = 1
				currentBidder = trialMain[1]["name"]
				minBid = minBidTMS
				SendChatMessage("Moving to trial main specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			elseif (trialOff[1]) then
				bidCategory = "to"
				bidderIndex = 1
				currentBidder = trialOff[1]["name"]
				minBid = minBidTOS
				SendChatMessage("Moving to trial off specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			else
				if (bMult) then SendChatMessage(item.." #"..curNum..": DE", channel)
				else SendChatMessage(item..": DE", channel) end
				ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
				if (itemCount > 1) then
					itemCount = itemCount - 1
					curNum = curNum + 1
					currentBidder = nil
					topBidder = nil
					ARP:StartBid()
				else
					ARP:Reset()
				end
			end
		elseif (bidCategory == "tm") then
			if (trialOff[1]) then
			bidCategory = "to"
				bidderIndex = 1
				currentBidder = trialOff[1]["name"]
				minBid = minBidTOS
				SendChatMessage("Moving to trial off specs! "..currentBidder.." min "..minBid, channel)
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			else
				if (bMult) then SendChatMessage(item.." #"..curNum..": DE", channel)
				else SendChatMessage(item..": DE", channel) end
				ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
				if (itemCount > 1) then
					itemCount = itemCount - 1
					curNum = curNum + 1
					currentBidder = nil
					topBidder = nil
					ARP:StartBid()
				else
					ARP:Reset()
				end
			end
		else
			if (bMult) then SendChatMessage(item.." #"..curNum..": DE", channel)
			else SendChatMessage(item..": DE", channel) end
			ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
			if (itemCount > 1) then
				itemCount = itemCount - 1
				curNum = curNum + 1
				currentBidder = nil
				topBidder = nil
				ARP:StartBid()
			else
				ARP:Reset()
			end
		end		
	else
		if (bidCategory == "mm" or bidCategory == "tm") then
			if (winAmt < 10) then winAmt = 10 end
		else
			if (winAmt < 1) then winAmt = 1 end
		end
		if (bMult) then SendChatMessage(item.." #"..curNum..": "..winner.." for "..winAmt, channel)
		else SendChatMessage(item..": "..winner.." for "..winAmt, channel) end
		ARP:SendMessage({ "WINNER", item, curNum, winner, winAmt, ARP:BuildLootLogText(winner, winAmt) })
		if (itemCount > 1) then
			itemCount = itemCount - 1
			curNum = curNum + 1
			currentBidder = nil
			topBidder = nil
			minBid = nil
			local equipLoc = select(9, GetItemInfo(item)) -- should already be cached since we checked the item when we were assigned it
			if (equipLoc ~= "INVTYPE_2HWEAPON" and equipLoc ~= "INVTYPE_WEAPONMAINHAND" and 
				equipLoc ~= "INVTYPE_WEAPONOFFHAND" and equipLoc ~= "INVTYPE_WEAPON") then
				ARP:RemoveWinner(winner)
			end
			ARP:StartBid()
		else 
			ARP:Reset() 
		end
	end
end

function ARP:RemoveWinner(winner)
	for i,v in pairs(ARP:GetList()) do
		if (v["name"] == winner) then
			table.remove(ARP:GetList(), i)
			break
		end
	end
end

function ARP:StartBid_UI()
	if bBidding then
		print("Your bid is already in progress!")
	else
		ARP:StartBid()
	end
end

function ARP:StartBid(skipCheck)
	if (not item) then print("You have not been assigned an item to bid!")
	else
		ARP:BuildBidString(item)
		if (not memberMain[1] and not memberOff[1] and not trialMain[1] and not trialOff[1]) then
			if (bMult) then 
				if (itemCount == 1) then 
					SendChatMessage(item.." #"..curNum..": DE", channel)
					ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
					ARP:Reset() 
					return
				end
				SendChatMessage(item.." #"..curNum..": DE", channel)
				itemCount = itemCount - 1
				curNum = curNum + 1
				ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
				ARP:StartBid()
			else 
				SendChatMessage(item..": DE", channel) 
				ARP:SendMessage({ "WINNER", item, curNum, nil, nil, ARP:BuildLootLogText() })
				ARP:Reset()
			end
		else
			ARP:BidItem(channel)
		end
	end
end

function ARP:UpdateBid(bidAmt)
	local list = ARP:GetList(bidCategory)
	local msg = ""
	if (bidAmt == -1) then
		if (bidderIndex == #list) then 
			SendChatMessage(item..": "..currentBidder.." has bid "..minBid.."! No more bidders.", channel)
			ARP:EndBid(currentBidder, minBid)
		else
			bidderIndex = bidderIndex + 1
			for i=bidderIndex, #list do
				if (i == #list) then msg = (msg..list[i]["name"])
				else msg = (msg..list[i]["name"]..", ") end
			end
			msg = (item..": "..currentBidder.." has bid "..minBid.."! Remaining bidders: "..msg)
			SendChatMessage(msg, channel)
			topBidder = currentBidder
			currentBidder = list[bidderIndex]["name"]
		end
	elseif (bidAmt == "pass" or bidAmt == "nope") then
		if (bidderIndex == #list) then
			SendChatMessage(item..": "..currentBidder.." has passed! No more bidders.", channel)
			ARP:EndBid(topBidder, minBid-1)
		else
			bidderIndex = bidderIndex + 1			
			for i=bidderIndex, #list do
				if (i == #list) then msg = (msg..list[i]["name"])
				else msg = (msg..list[i]["name"]..", ") end
			end
			if (topBidder == nil) then topBidder = "None" end
			msg = (item..": "..currentBidder.." has passed! Top bidder: "..topBidder.."; Remaining bidders: "..msg)
			SendChatMessage(msg, channel)
			currentBidder = list[bidderIndex]["name"]
			SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
		end
	else
		if (string.sub(bidAmt, 1, 1) == "+") then bidAmt = ((minBid-1) + tonumber(bidAmt))
		elseif (bidAmt == "yup") then bidAmt = minBid
		else bidAmt = tonumber(bidAmt) end
		if (bidAmt < minBid) then
			SendChatMessage("Your bid is too low! The minimum is "..minBid..".", "WHISPER", nil, currentBidder)
		elseif (bidAmt > list[bidderIndex]["points"]) then
			SendChatMessage("You don't have that many points! You have "..list[bidderIndex]["points"]..".", "WHISPER", nil, currentBidder)
		else
			topBidder = currentBidder
			if (bidderIndex == #list) then
				if (bidAmt > minBid) then bidAmt = minBid end
				SendChatMessage(item..": "..currentBidder.." has bid "..bidAmt..". No more bidders!", channel)
				ARP:EndBid(topBidder, bidAmt)
			else
				minBid = bidAmt + 1
				bidderIndex = bidderIndex + 1
				for i=bidderIndex, #list do
					if (i == #list) then msg = (msg..list[i]["name"])
					else msg = (msg..list[i]["name"]..", ") end
				end
				msg = (item..": "..currentBidder.." has bid "..bidAmt..". Remaining bidders: "..msg)
				SendChatMessage(msg, channel)
				currentBidder = list[bidderIndex]["name"]
				SendChatMessage("It is your bid! Min "..minBid..".", "WHISPER", nil, currentBidder)
			end
		end
	end
end

function ARP:RemoveBidder(name)
	for i,v in pairs(memberMain) do
		if (v["name"] == name) then
			if (UnitName("player") == name) then
				print("You have been removed from the bid list for "..item..".")
			else
				SendChatMessage("You have been removed from the bid list for "..item..".", "WHISPER", nil, name)
			end
			table.remove(memberMain, i)
			return
		end
	end
	
	for i,v in pairs(memberOff) do
		if (v["name"] == name) then
			if (UnitName("player") == name) then
				print("You have been removed from the bid list for "..item..".")
			else
				SendChatMessage("You have been removed from the bid list for "..item..".", "WHISPER", nil, name)
			end
			table.remove(memberOff, i)
			return
		end
	end
	
	for i,v in pairs(trialMain) do
		if (v["name"] == name) then
			if (UnitName("player") == name) then
				print("You have been removed from the bid list for "..item..".")
			else
				SendChatMessage("You have been removed from the bid list for "..item..".", "WHISPER", nil, name)
			end
			table.remove(trialMain, i)
			return
		end
	end
	
	for i,v in pairs(trialOff) do
		if (v["name"] == name) then
			if (UnitName("player") == name) then
				print("You have been removed from the bid list for "..item..".")
			else
				SendChatMessage("You have been removed from the bid list for "..item..".", "WHISPER", nil, name)
			end
			table.remove(trialOff, i)
			return
		end
	end
	
	print("Could not find player to remove.")
end

function ARP:UpdateList(list, name, points, other_list)
	local replacing = false
	local found = nil
	local randnum = nil
	
	for i,v in ipairs(other_list) do
		if (v["name"] == name) then
			found = i
		end
	end
	if (found ~= nil) then table.remove(other_list, found) end
		
	for i,v in ipairs(list) do
		if (v["name"] == name) then
			v["points"] = points
			replacing = true
			break
		end
	end
	
	if (not replacing) then 
		randnum = randnums[name]
		if randnum == nil then
			randnum = math.random() 
			randnums[name] = randnum
		end
		table.insert(list, {name = name, points = points, randnum = randnum})
	end	
end

function ARP:GetList()
	if (bidCategory == "mm") then return memberMain
	elseif (bidCategory == "mo") then return memberOff
	elseif (bidCategory == "tm") then return trialMain
	elseif (bidCategory == "to") then return trialOff
	end
end

function ARP:Reset()
	memberMain = {}
	memberOff = {}
	trialMain = {}
	trialOff = {}
	randnums = {}
	
	currentBidder = nil
	bidderIndex = nil
	topBidder = nil
	item = nil
	itemCount = 1
	curNum = 1
	
	bBidding = false
	bMult = false
	ARP:SendMessage({ "READY" })
	ARP:MLPanel_Update()
end

function ARP:BidSort(list)
	table.sort(list, function(a, b)
		if a["points"] == b["points"] then 
			return a["randnum"] < b["randnum"] 
		else 	
			return a["points"] < b["points"]
		end
	end)
	return list
end

function ARP:GetLootLogLevelString(ilvl)
	if (ilvl == nil) then -- just in case...
		return " (???)"
	end
	for i,v in pairs(heroicLoot) do
		if (ilvl == v) then return " (H)" end
	end
	for i,v in pairs(normalLoot) do
		if (ilvl == v) then return "" end
	end
	return " ("..ilvl..")"
end

function ARP:BuildLootLogText(winner, winAmt)
	local itemName, _, _, ilvl = GetItemInfo(item) -- should already be cached since we checked the item when we were assigned it
	
	if (itemName == nil) then -- just in case...
		itemName = "???"
	end

	local playerName
	if winner == nil then playerName = "DE" else playerName = winner end

	local winningBidAmount = ""
	if winner ~= nil then winningBidAmount = " " .. winAmt end
	
	return itemName .. ARP:GetLootLogLevelString(ilvl) .. " = " .. playerName .. winningBidAmount
end